# =====================================================================
# csaCgi.sed: CGI API pre-processor for CSA.cgi.
#
# Copyright (c) 2008,2009 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================

# First input line is the value of SCRIPT_NAME
1{
	s,/CSA\([^/]\)\?.*,/CSA\1,
	h
	d
}

# Second input line is the value of PATH_INFO
2{
	# Make sure value ends with a single slash.
	s,/*$,/,

	# Set bogus application name if empty.
	s,^/$,/__BOGUS__/,

	# If no CSA personality flag is present, set the bogus one '"' .
	s,^\(/[^/]\+/\)\([^IRS2P][^/]\),\1"/\2,

	# Handle special case.
	s,^\(/[^/]\+/\)$,\1"/,

	/^\/[^\/]\+\/[IRS2P"]\//{
		H
		s,\(/[^/]\+/\)[IRS2P"]/,\1,
		H

		# Strip leading CSA application name from PATH_INFO.
		#s,^/[^/]\+/,/,

		# Print rewritten PATH_INFO. If it ends up in containing
		# a single '/' it is up to the calling program to consider
		# it empty, but we must not make it completely empty here,
		# as the resulting arg list must not have nil values in
		# positions 1 and 2.
		s,/*$,,
		s,^$,/,
		p

		g
		s,\n/[^/]\+/\([IRS2P"]\)/.*,\1,

		# Print rewritten SCRIPT_NAME, after stripping trailing '"'
		s,",,
		p

		# Print rewritten CSA_PGM, after stripping the trailing '/'
		s,.*/,,
		p
	}
}

# Third line is target method name, received through QUERY_STRING,
# and it must be turned into the canonical CSA format 0=app.method .
# No test is done to check whether the value is already canonical,
# as the method name is allowed to contain dots itself so I have
# no way to assess wheter a dot stands for an app.method separator or
# it is rather a method name with dots in it. So, if an application
# name is available in PATH_INFO I use it in any case.

3{
	/^0=/{
		x
		s,.*\n,0=,
		s,/,,
		s,/,.,g
		s,\(0=[^.]\+\),\1,

		x
		s,^0=,,
		s,^,|,
		H

		g
		s,\n|,,
		s,__BOGUS__\.,,
		p
	}
}

# Fourth line is the value of CONTENT_TYPE, which MAY be specified
# by the client appended with the charset being used, like this:
#
# "application/x-www-form-urlencoded; charset=UTF-8"
#
# If that is the case then I need to extract the charset spec
# and use it to set the initial value of CSA_REQ_CHARSET .
# Add more acceptable charsets as needed.

4{
	y/ABCDEFGHIJKLMNOPQRSTUVWXYZ/abcdefghijklmnopqrstuvwxyz/

	s,.* charset=,,

	/^utf-8/p

	# No need to extract other charsets, but leave as an example.
	#/^iso-8859-1/p
	#/^iso-8859-15/p
}

#EOF
