# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2001,2006 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# float _julian(int year, int month, int day \
#			[, int hours, int minutes, int seconds])
#
# Convert calendar to Julian date (Julian day number algorithm adopted
# from Press et al.).
#
# Notes:
# CE = Common Era (AD), i.e. dates after 1582-10-14
# BCE = Before Common Era, i.e. dates before 1582-10-04 .
# =====================================================================

function _julian(y,m,d,h,mn,s,		jy,ja,jm,intgr,gregcal,frac, \
					dayfrac,jd,jd0,era) {

   # Make sure args cast to numbers, or rounding errors may occur.
   y += 0; m += 0; d += 0; h += 0; mn += 0; s += 0

   if (m < 1 || m > 12 || d < 1 || d > 31 || \
	h < 0 || h > 23 || mn < 0 || mn > 59 || s < 0 || s > 59)
							return _NULL
   if (y == 0) {
      print "_julian: there is no year 0 in the Julian system!" > __STDERR
      return _NULL
   }

   if (y == 1582 && m == 10 && d > 4 && d < 15) {
      print "_julian: the dates 5 through 14 October, 1582, " \
	    "do not exist in the Gregorian system!" > __STDERR
      return _NULL
   }

   if (y < 1582) era = "BCE"
   else if (y > 1582) era = "CE"
   else if (m < 10) era = "BCE"
   else if (m > 10) era = "CE"
   else if (d < 5) era = "BCE"
   else if (d > 14) era = "CE"
   else era = "CE"

#  if (y < 0) ++y
   if (era == "BCE") y = -y + 1
   if (m > 2) {
      jy = y
      jm = m + 1
   }  else {
      jy = y - 1
      jm = m + 13
   }

   intgr = int(int(365.25*jy) + int(30.6001*jm) + d + 1720995)

   # check for switch to Gregorian calendar
   gregcal = 15 + 31*(10 + 12*1582)
   if (d + 31*(m + 12*y) >= gregcal) {
      ja = int(0.01*jy)
      intgr += 2 - ja + int(0.25*ja)
   }

   # correct for half-day offset
   dayfrac = h/24.0 - 0.5
   if (dayfrac < 0.0) {
      dayfrac += 1.0
      --intgr
   }

   # now set the fraction of a day
   frac = dayfrac + (mn + s/60.0)/60.0/24.0

   # round to nearest second
   jd0 = (intgr + frac)*100000
   jd  = int(jd0)
   if (jd0 - jd > 0.5) ++jd
   return sprintf("%.5f",jd/100000)
}
