# *********************************************************************
#  Written by and copyright Carlo Strozzi <carlos@linux.it>.
# 
#  post2rc.sed: decode HTTP POST form variables.
#  Copyright (C) 2002 Carlo Strozzi <carlos@linux.it>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# *********************************************************************
#:Begin1
#:Name       post2rc.sed
#:File       lib/post2rc.sed
#:Summary    sed(1) program file
#:Project    CSA
#:Layer      CSA
#:Operation  Decode HTTP POST form variables and turns them into valid
#:	     rc(1) assignments. This is a sed(1) re-implementation of
#:	     the "uncgi" utility. See uncgi home page for more info:
#:	     http://www.midwinter.com/~koreth/uncgi.html
#:Author     Carlo Strozzi <carlos@linux.it>
#:Maintainer Carlo Strozzi <carlos@linux.it>
#:Revision   $Date: 2003/03/17 11:37:29 $
#:Platforms  Linux
#:Copying    GPL
#:Bugs	     sed(1) is uncomparably slower and awkward than AWK at these
#:	     things. So what's the point ? Well, it can be great fun :-)
#:	     Depending on the amount of input, this program may take
#:	     from 100ms to even over four seconds on a P-II 233MHz.
#:	     Multi-valued variables (i.e. "select multiple" form elements)
#:	     are not yet supported.
#:End
#
# ChangeLog:
#
# * Development code
#   2002-02-14 Carlo Strozzi <carlos@linux.it>
#
# *********************************************************************

# truncate input stream to 10 KB to prevent buffer and
# environment overflows in the calling program
s/^\(.\{10000\}\).*$/\1/

# Quit if input contains no assignments. This is either an ISINDEX
# query -- in which case this program should not have been called in
# the first place -- or garbled input from a broken client.
/=/!d

# remove trailing assignments that have been invalidated by
# the truncation process
s/&[^&=]*$//
s/%.\?$//

# single-quotes must be doubled for rc(1)
s/%27/''/g

# insert semicolon at the beginning
s/^/;/

# add closing-quote and semicolon at the end
s/$/';/

# split-up assignments for rc(1) and add closing quotes
s/&/';/g

# Handle "select multiple" form elements in an uncgi-like fashion.
# This is currently not implemented yet.

#/;\([^=]*\)=\([^;]*\);\(.*\)\1/{
#	s/$/ ..... etc ...
#}

# replace invalid characters with underscores (_) in variable names
:1
s/;\([A-Za-z0-9_]*\)\([^A-Za-z0-9_=]\)\([^=]*\)=/;\1_\3=/g
t1

# insert opening quotes
s/=/='/g

# normalize line-breaks a-la uncgi

# turn CR+LF into LF in values which name begins with "_"
:2
s/\(;_[^;]*=\)\([^;]*\)%0D%0A\([^;]*;\)/\1\2%0A\3/
t2

# turn CR into LF in values which name begins with "_"
:3
s/\(;_[^;]*=\)\([^;]*\)%0D\([^;]*;\)/\1\2%0A\3/
t3

# remove leading "_" from names
s/;_\([^;]*;\)/;\1/g

# prefix all names with WWW_
s/;\([^$]\)/;WWW_\1/g

# remove leading ";"
s/^;//

# convert lower-case escapes into upper-case ones (i.e %f5 -> %F5)

# add lookup table to the end
s/$/\
aAbBcCdDeEfF/

# perform case-conversion on the first digit.
:4
s/\(%\)\([a-f]\)\(.*\)\n\(.*\)\2\(.\)\(.*\)/\1\5\3\
\4\2\5\6/
t4

# perform case-conversion on the second digit.
:5
s/\(%.\)\([a-f]\)\(.*\)\n\(.*\)\2\(.\)\(.*\)/\1\5\3\
\4\2\5\6/
t5

# discard the lookup table
s/\n.*//

# Append a new lookup table. For the sake of speed, I use only a (large)
# subset of all the possible ascii and iso-8859-1 escape codes. What has
# been left out are codes for numbers, upper- and lower- case letters,
# which seems reasonable. Unfortunately the overall process remains
# quite slow anyway, especially with GNU sed.

s/$/\
%07%08%09	%0B%0C%20 %21!%22"%23#%24$%26\&%28(%29)%2A*%2C,%2D-%2E.%2F\/%3A:%3B;%3C<%3D=%3E>%3F?%40@%5B[%5C\\%5D]%5E^%5F_%60`%7B{%7C|%7D}%7E~%A0 %A1%A2%A3%A4%A5%A6%A7%A8%A9%AA%AB%AC%AD%AE%AF%B0%B1%B2%B3%B4%B5%B6%B7%B8%B9%BA%BB%BC%BD%BE%BF%C0%C1%C2%C3%C4%C5%C6%C7%C8%C9%CA%CB%CC%CD%CE%CF%D0%D1%D2%D3%D4%D5%D6%D7%D8%D9%DA%DB%DC%DD%DE%DF%E0%E1%E2%E3%E4%E5%E6%E7%E8%E9%EA%EB%EC%ED%EE%EF%F0%F1%F2%F3%F4%F5%F6%F7%F8%F9%FA%FB%FC%FD%FE%FF/

# Lookup the escaped value and replace it with the unescaped one, if found.
# The replacement order does not matter here.
# Lookup misses will remain unescaped.

:6
s/\(%..\)\(.*\)\n\(.*\)\1\(.\)\(.*\)/\4\2\
\3\1\4\5/
t6

# discard the lookup table
s/\n.*//

# perform order-dependent substitutions
s/+/ /g
s/%2B/+/g
s/%25/%/g

# this must come last
s/%0A/\
/g

# in case a broken client sends input on multiple lines
1q

# End of program
