# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2001,2006,2007,2008,2009 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# boolean _isuri(string uri [,boolean url])
#
# Return _TRUE if the received string is a sintactically valid URI, or
# _FALSE otherwise. If option 'url' is set to '1' then the check will
# return _TRUE only if "uri" is in fact a URL, which syntax is a subset
# of the more general URI class. The test for URIs will spot also
# CURIE type URIs, see http://en.wikipedia.org/wiki/CURIE .
# =====================================================================

function _isuri(uri,url) {

  # This pattern matches URIs in a very broad manner, but on the other
  # hand it is rather difficult to be more selective. Furthermore, it
  # does not currently check whether the received string contains
  # invalid URI characters, beside SPACE, '\t', '\r', and '\n'.

  if (url) {
     if (uri ~ /^[a-z][a-z0-9]+:\/\/[^\/]+[^ \t\r\n]+$/) return _TRUE
  }

  # Detect also CURIEs and Blank Node (bnode) CURIEs.
  # See http://www.w3.org/MarkUp/2009/rdfa-for-html-authors .

  else if (uri ~ /^[a-z][a-z0-9]+:[^] \t\r\n\[]+$/ || \
	uri ~ /^\[[a-z][a-z0-9]+:[^] \t\r\n\[]+]$/ || \
	uri ~ /^\[_:[^] \t\r\n\[]+]$/) return _TRUE

  return _FALSE
}
