# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2004,2007 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# anytype _request(string name [,int element, boolean opt])
#
# Read the value of 'name' from the list of name/value pairs received
# through the CGI interface. Non-CGI and CSA CGI special names can be
# assessed by looking them up in ENVIRON[] directly. Since 'name' is
# usually an rc(1) list of one or more values, if 'element' is not
# specified, or it is '0', then then the whole list is returned, in the
# form of a scalar, possibly with embedded \001's; if 'element' is > 0
# then the corresponding element, if available, is returned; if 'element'
# is < 0 then the last element is returned. If the requested name and/or
# element do not exist, then _ERRNO is set and a null value is returned.
# This is to enable the caller to cope with variable-length argument
# lists simply by doing something like:
#
#   var = _request("arg1",1)
#   if (_ERRNO == _EINVAL) print "argument arg1.1 does not exist!"
#
# If 'opt' is '1' then 'name' is considered to be an uploaded object
# and the value of its 'filename=' attribute is returned, if available.
#
# If 'opt' is '2' then 'name' is considered to be an uploaded object
# and it is checked for a safe path. If the path is found to be not
# safe then a safe /dev/null value is returned.
# =====================================================================

function _request(name,element,opt,			a,i,j,value) {

  element += 0				# cast numeric.

  value = ENVIRON["WWW_" name]

  if (opt == _O_UPNAME || opt == _O_ATTACH) {

     i = split(value, a, "\001")

     value = _NULL

     for (j=1; j<=i; j++) {

	 # detect unsafe attachment path and make it inoffensive.
	 if (a[j] ~ /\.\.\// || \
	     a[j] !~ /^\/.*hexBinary\.[^\/]+\.tmp$/) a[j] = "/dev/null"

	 # the path was safe, but does it really refer to an uploaded
	 # object ? Better take the belt-and-braces approach ;-)

	 else if (_UPLOAD[a[j]] != a[j]) a[j] = "/dev/null"

	 if (j > 1) value = value "\001"

	 if (opt == _O_UPNAME) {
	    if (a[j] == "/dev/null") a[j] = _NULL
	    else {
	       sub(/.*\/hexBinary\.[0-9]+\.file\./,_NULL,a[j])
	       sub(/\.arg\.[^.]+\.tmp$/,_NULL,a[j])
	    }
	    value = value _uridecode(a[j])
	 }
	 else value = value a[j]
     }
  }

  if (element == 0) return value
  if ((i=split(value,a,"\001")) == 0 || element > i) _ERRNO = _EINVAL
  else _ERRNO = 0			# Always reset if not needed!
  if (element < 0) return a[i]
  else return a[element]
}

