# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2001,2008 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# string _xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001 $[
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appear anywhere in a document, not even as character references, such
# as &#31;. The only other possibility it to either remove them or turn
# them into something sensible. I took this second approach, whereby
# \001 (that is rc(1) list concatenation character) is replaced with a
# single blank. This is experimental, and may change in future versions
# of CSA.
# =====================================================================

function _xmlencode(string) {

   gsub(/&#36;\[/,"$[",string)			# must come very first.
   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/]]>/, "]]\\&gt;", string)		# Forbidden sequence
   gsub(/"/, "\\&quot;", string)		# double quote

   # Single quote (MSIE 5/6 does not expand "&apos;". Unfortunately,
   # this test does not fix the problem for pages that can be edited
   # online, such as Wiki pages, as they will use either '&apos;' or
   # '&#39;' depending on which broser was used to edit them. The safest
   # bet is therefore to always use '&#39;'.

   #if (ENVIRON["HTTP_USER_AGENT"] ~ / MSIE /) gsub(/'/,"\\&#39;",string)
   #else gsub(/'/, "\\&apos;", string)

   gsub(/'/, "\\&#39;", string)

   gsub(/\001/, " ", string)			# rc(1)'s ^A

   # This is MANDATORY to prevent users from injecting CSA tags.
   gsub(/\$\[/,"\\&#36;[",string)

   return string
}

