#!/usr/bin/rc -p
# *********************************************************************
# CSA.cgi: CGI Shell Architecture Remote Procedure Call processor.
#
# Copyright (c) 2002,2009 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# *********************************************************************

# *********************************************************************
#			Local functions
#
# Note: to save environment space, functions have the same name as the
# actual CSA library ones. This may not matter if rc(1) subprograms
# are called with flag '-p', but it definitely does matter if they are
# included with ``. file''.
# *********************************************************************

# Define a few dummy functions that may be called by this script before
# their real versions have been loaded.

fn csaAuth.send csaSetTime csaUnlist csaUniq {return 0}

# Although no csaDebug() is supposed to be called before the underlying
# application layer is initialized, I nevertheless define a simplified
# debug function to try and catch really everything since the very
# early stages of a CSA run, just in case.

fn csaDebug {
   CSA_DEBUG_STEP = ($CSA_DEBUG_STEP 1)

   ~ $#CSA_DEBUG_STEP 1 &&
     echo $0 $REQUEST_METHOD $PATH_INFO $QUERY_STRING >[1=2]

   echo CSA.cgi[$pid,NO_RID_YET,$#CSA_DEBUG_STEP] $* >[1=2]
}

# =====================================================================
# csaTrue value
# 
# Test whether 'value' has a value of either 'true', '1' or 'y[es]',
# with all values case-insensitive.
# =====================================================================

fn csaTrue {

   ~ $CSA_DEBUG 0 || csaDebug $0 $*

   ~ $1 1 [Tt][Rr][Uu][Ee] [Yy] [Yy][Ee][Ss] && return 0

   return 1
}

fn csaCleanup {
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*

   #csaUnlist CSA_WORKFILES $CSA_RPC_CMD $CSA_RPC_AWK \
   #					$CSA_RPC_PUT $CSA_RDF_ARG

   # Prevent wasting time trying to remove the same files twice,
   # and remove any reference to /dev/null, to prevent wasteful
   # cluttering of stderr.log. Note: CSA_LOCK_SEMAPHORE must
   # be removed *after* CSA_WORKFILES.

   csa1 = ($CSA_RPC_CMD $CSA_RPC_AWK $CSA_RPC_PUT
	   $CSA_RDF_ARG $CSA_WORKFILES $CSA_CGI_STDIN) csa2=() {

	# Uploaded files may be referenced inside commit lists, so
	# they should be trapped only if workfiles are not being
	# retained as a consequence of commit errors, or an attempt
	# to re-run 'csa-db-repair' may no longer have all it needs
	# to be repeatable.

	~ $CSA_WORKFILES () ||
	  CSA_WORKFILES = ($CSA_WORKFILES $TMPDIR/hexBinary.$pid.*)

	csaUniq csa1; csaUnlist csa1 /dev/null

	if (~ $CSA_LOCK_SEMAPHORE /*) {
	   csaTrue $CSA_LOCK_VERBOSE && csa2 = --verbose
	   if (!~ $CSA_CMD_DOTLOCK ()) {
	      ~ $csa2 --verbose && !~ $CSA_LOCKS () &&
		echo 'dot-lock: removing lockfile(s)' $CSA_LOCKS >[1=2]
	      echo 'rm -rf -- $csa1 $CSA_LOCK_SEMAPHORE $CSA_LOCKS' > $CSA_LOCK_SEMAPHORE
	   }  else echo 'rm -rf -- $csa1 $CSA_LOCK_SEMAPHORE; ~ $CSA_LOCKS () || swu-unlock $csa2 $CSA_LOCK_HOST $CSA_LOCKS' > $CSA_LOCK_SEMAPHORE
	   . $CSA_LOCK_SEMAPHORE
	}  else rm -rf -- $csa1
   }
}

fn csaExit {
   # Preserve workfiles if requested. This must be handled first,
   # even before csaDebug is called.
   if (~ $1 -n --no-cleanup) {
      CSA_WORKFILES = ()
      shift
   }
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*
   csaCleanup
   exit $1
}

fn csaHttp.header {
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*
   csa1=() csa2=() {
     while (~ $1 -*) {
	switch ($1) {
	  case -e --end-of-header
	       csa1=$nl
	  case -d --deferred
	       csa2=1
	}
	shift
     }
     ~ $#* 2 || return 0			# nothing to do.
     ~ $1 [Ee]xpires && ~ $2 -1 &&
	  *=($1 'Thu, 01 Jan 1970 00:00:00 +0000')
     if (~ $csa2 ()) {
     	echo $1:' '$2$csa1
     }	else {
	~ $CSA_HTTP_HEADERS () || CSA_HTTP_HEADERS=$CSA_HTTP_HEADERS$nl
	CSA_HTTP_HEADERS=$CSA_HTTP_HEADERS$1:' '$2$csa1
     }
   }
}

# *********************************************************************
# HTTP/1.1 Status Codes:
#
# 1xx Informational:
#
# 100-199: Codes in the 100s are informational, indicating that the
#	   client should respond with some other action.
#
#
# 2xx Success:
#
# 200-299: Values in the 200s signify that the request was successful.
#
#
# 3xx Redirection:
#
# 300-399: Values in the 300s are used for files that have moved
#	   and usually include a Location header indicating the new
#	   address.
#
#
# 4xx Client Error:
#
# 400-499: Values in the 400s indicate an error by the client.
#
#
# 5xx Server Error:
#
# 500-599: Codes in the 500s signify an error by the server.
#
#
# See http://www.informit.com/articles/article.asp?p=29817&seqNum=7
# (among other places) for more information.
#
# See also http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
#
# *********************************************************************

fn csaHttp.status {
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*

   csa1=() {

     # Add more NPH auto-detection here.
     switch ($SCRIPT_NAME) {
       case /cgi-bin/nph-*
	    csa1 = nph
     }

     ~ $1 [1-5][0-9][0-9] || * = ()
     switch ($1) {
       case 200; * = (7  $1 OK)
       case 201; * = (12 $1 Created)
       case 202; * = (13 $1 Accepted)
       case 204; * = (0  $1 No Content)
       case 301; * = (22 $1 Moved Permanently)
       case 302; * = (22 $1 Moved Temporarily)
       case 303; * = (14 $1 See Other)
       case 307; * = (23 $1 Temporary Redirect)
       case 400; * = (16 $1 Bad Request)
       case 401; * = (17 $1 Unauthorized)
       case 403; * = (14 $1 Forbidden)
       case 404; * = (14 $1 Not Found)
       case 405; * = (23 $1 Method Not Allowed)
       case 406; * = (19 $1 Not Acceptable)
       case 409; * = (13 $1 Conflict)
       case 413; * = (29 $1 Request Entity Too Large)
       case 500; * = (26 $1 Internal Server Error)
       case 501; * = (20 $1 Not Implemented)
       case 503; * = (24 $1 Service Unavailable)
       case *;   * = (23 550 CSA Internal Error)	# Unknown status code.
     }

     CSA_HTTP_LENGTH = $1; shift
     CSA_HTTP_STATUS = $*

     if (~ $csa1 ()) {
        csaHttp.header Status $^CSA_HTTP_STATUS
     }	else {
	echo HTTP/1.0 $^CSA_HTTP_STATUS
	csaHttp.header Date $CSA_TIME_HTTP
	csaHttp.header Server $SERVER_SOFTWARE
	csaHttp.header X-Parsed no		# useful for debugging.
     }
   }
}

fn csaExit.http {
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*
   csaCleanup
   csaHttp.status $1

   # Experiments have shown that it is preferable not to forbid
   # caching on the part of the client, but rather tell the
   # latter that the document has changed. 

   csaHttp.header Last-Modified $CSA_TIME_HTTP

   # The following cache-related directives do more harm than good,
   # so don't use them by default.
   #
   #csaHttp.header Pragma no-cache			# for HTTP/1.0
   #csaHttp.header Cache-control no-cache		# for HTTP/1.1
   #csaHttp.header Expires -1

   csaHttp.header X-Powered-By CSA/$CSA_VERSION
   csaAuth.send					# mandatory!

   ~ $CSA_HTTP_LENGTH 0 || csaHttp.header Content-Length $CSA_HTTP_LENGTH

   # Make picky clients happy.
   csaHttp.header --end-of-header Content-Type text/plain

   # Add more "silent" codes if necessary.
   ~ $CSA_HTTP_STATUS(1) 204 || echo $^CSA_HTTP_STATUS
   exit
}

fn csaExit.fault {
   # Preserve workfiles if requested. This must be handled first,
   # even before csaDebug is called.
   if (~ $1 -n --no-cleanup) {
      CSA_WORKFILES = ()
      shift
   }
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*
   ~ $1 [45][0-9][0-9] || * = 550
   csaExit.http $1
}

fn csaExit.ok {
   ~ $CSA_DEBUG 0 false || csaDebug $0 $*
   csaExit.http 204
}

# *********************************************************************
#                This stuff is needed very early
# *********************************************************************

# This is needed very early.
nl='
'

# Make sure any locale is reset to POSIX, otherwise the regular
# expression [a-zA-Z] will match also accented letters with gawk(1)
# and GNU sed(1).
LANG = ()

# Load local settings.
. '@CSA_INSTALL@/etc/csa.rc'

# Set a sane path.
# This can only be done after the global csa.rc is sourced.
path = (($SWU_INSTALL $CSA_INSTALL /usr/local /usr)^/bin /bin)

# *********************************************************************
#                Start of user-modifiable settings
# *********************************************************************

# Adjust filtering as desired and set this switch to '1' to trace
# really everything, see below.
~ $REMOTE_ADDR * && CSA_DUMP = 0

# Change these settings as appropriate and then copy this script
# to your HTTP daemon /cgi-bin/ directory. If you are using per-user
# /cgi-bin/ areas (usually with the Apache Web Server SuEXEC or with
# the cgiwrap utility), then act as appropriate, according to the
# relevant documentation of said techniques.

~ $logname () && logname = `{whoami}
~ $home () && home = /home/$logname
~ $libstem () && libstem = $home/.csa
~ $tmpdir () && tmpdir = /tmp

# *********************************************************************
#                 End of user-modifiable settings
# *********************************************************************

# Do not rewrite things multiple times.

if (!~ $CSA_DUMP_FILE /*) {

   # If either XML-RPC or SOAP, automatically set the pseudo method
   # in QUERY_STRING if not specified by the client. By doing this,
   # the client will be able to invoke the RPC peronalities of CSA
   # simply as /cgi-bin/CSA/<class>/[2P] (*optionally* followed by
   # a QUERY_STRING), rather than having to use one of the more
   # verbose canonical methods such as /cgi-bin/CSA[2P]?0=<class>.RPC
   # and /cgi-bin/CSA[2P]?0=RPC (also optionally followed by further
   # QUERY_STRING arguments). This code has been tested and it is
   # known to work on both thttpd(8) and apache(8).

   isindex = $1				# save value for later.

   * = ``(/$nl){echo $PATH_INFO}
   if (~ $2 2 P && !~ $QUERY_STRING '0='^*) {
      if (~ $QUERY_STRING ()) {
	 QUERY_STRING = '0=RPC'
      }  else QUERY_STRING = '0=RPC&'$QUERY_STRING
   }

   # Account for the fact that QUERY_STRING and PATH_INFO can be either
   # undefined or nil, depending apparently on which web server we are.

   # Needed by csaCgi.awk.
   if (~ $PATH_INFO^X X) {
      CSA_REQUEST_URI = /UNKNOWN
      PATH_INFO = ()				# make it really null.
   }  else CSA_REQUEST_URI = $PATH_INFO

   if (~ $QUERY_STRING^X X) {
      CSA_REQUEST_URI = $CSA_REQUEST_URI'?UNKNOWN'
      QUERY_STRING = ()				# make it really null.
   }  else CSA_REQUEST_URI = $CSA_REQUEST_URI'?'$QUERY_STRING

   * = ``$nl{echo \
	$SCRIPT_NAME$nl$PATH_INFO$nl$QUERY_STRING$nl$CONTENT_TYPE |
				sed -n -f $CSA_INSTALL/lib/csaCgi.sed}

   ~ $bqstatus 0 || csaExit.fault

   PATH_INFO = $1; ~ $PATH_INFO / && PATH_INFO = ()
   SCRIPT_NAME = $2; ~ $SCRIPT_NAME */CSA* || csaExit.fault
   CSA_PGM = $3

   switch ($5) {
     case utf-8; CSA_REQ_CHARSET = $5
     case iso-8859-*; CSA_REQ_CHARSET = iso-8859-1
   }

   # Override QUERY_STRING only if the relevant arg is non-empty.
   # Furthermore, if the client called us using the form '0=...'
   # then the sole purpose of PATH_INFO is to make selective
   # path-based cookies work, and once it has done its job it
   # must be dropped.

   ~ $4 '0='^* && QUERY_STRING = $4 && PATH_INFO = ()

   * = $isindex				# restore saved value.
}

~ $CSA_PGM () && CSA_PGM = $0; sessid = ()

if (~ $CSA_DUMP 1 && !~ $CSA_DUMP_FILE /*) {
   CSA_DUMP_FILE = `{mktemp /tmp/csa.stderr.$pid.XXXXXX}
   ~ $bqstatus 0 || exit 1
   CSA_STDERR = $CSA_DUMP_FILE
   NOSQL_STDERR = $CSA_DUMP_FILE
   NOSQL_DEBUG = 1
   CSA_ENV_FILE = `{mktemp /tmp/csa.env.$pid.XXXXXX}
   ~ $bqstatus 0 || exit 1
   rc -p -c env > $CSA_ENV_FILE
   CSA_ARGS_FILE = `{mktemp /tmp/csa.args.$pid.XXXXXX}
   ~ $bqstatus 0 || exit 1
   echo $* > $CSA_ARGS_FILE
   if (~ $REQUEST_METHOD POST PUT) {
      CSA_INPUT_FILE = `{mktemp /tmp/csa.stdin.$pid.XXXXXX}
      ~ $bqstatus 0 || exit 1

      # Truncating any input to the supplied length is always advisable,
      # or the Web server may hang until timeout.

      if (~ $CONTENT_LENGTH [1-9]*) {
	 head -c $CONTENT_LENGTH > $CSA_INPUT_FILE
      }	 else cat > $CSA_INPUT_FILE

      # append mandatory, or rc(1) will also dump garbage into the file!
      exec rc -p -x $0 $* < $CSA_INPUT_FILE >>[2]$CSA_DUMP_FILE
   }  else exec rc -p -x $0 $* >>[2]$CSA_DUMP_FILE
}

# Misc. defs

if (!~ $CSA_DUMP_FILE /*) {
   CSA_DEBUG=0					# Must be not-null.
}  else CSA_DEBUG=1

# The actual temporary directory will *not* be publicly writable !!
TMPDIR = $tmpdir/$logname-csa-$CSA_VERSION

CSA_RPC_UPLOAD = ()				# Must be null here.

# *********************************************************************
#			   Main program
# *********************************************************************

# Set time stuff, leaving any ISINDEX argument alone.

* = () {
  * = `{date '+%Y %m %d %H %M %S %Z %a %b %s %z %:z'}

  # Save values for possible later use by application programs
  # to built custom local date formats.
  CSA_TIME_LOCAL = $*

  CSA_TIME_ISO = $1-$2-$3' '$4:$5:$6

  # I set this to be in local-time only.
  # See http://www.w3.org/TR/NOTE-datetime for more information.
  CSA_TIME_ISO8601 = $1-$2-$3^T$4:$5:$6$12

  CSA_TIME_LOG = $CSA_TIME_ISO.$7

  ~ $7 UTC GMT && CSA_TIME_ISO = $CSA_TIME_ISO^Z

  # A second date(1) run is necessary to set UTC time values.
  * = `{date -u '+%Y %m %d %H %M %S %Z %a %b %s %z %:z'}

  # Save values for possible later use by application programs
  # to built custom UTC date formats.
  CSA_TIME_UTC = $*

  CSA_TIME_HTTP = $8,' '$3' '$9' '$1' '$4:$5:$6' 'GMT
}

# Set the program name. If more special names are added remember to add
# them also to csaMainlib.rc. Note that although CSA_RPC_FMT is set
# here, a CSA application may want to override it with a fixed value,
# typically "html", either in the application profile or in any other
# shell stuff that is called prior to sending the response to the client.
# This is the case, for instance, of a CSA application which templates
# are not well-formed XML so that they should not be used to convey any
# embedded RDF/RSS content. See also the use of CSA_RPC_FMT in csaCgilib.rc.
# Generally speaking, specific call formats imply specific response
# formats, while the opposite is not necessarily true. An exception to
# this general rule is the RSS 1.0 format, for which it is not enough
# to trigger the proper CSA_RPC_FMT setting, but provisions must be taken
# also at the _userproc(_O_RESPONSE) level, or it will not work.

switch ($CSA_PGM) {
  case *CSAI *CSAI.* *CSAI-*		# interactive response mode
       CSA_PGM = CSAI
  case *CSAR *CSAR.* *CSAR-*		# RDF response mode
       CSA_PGM = CSA
       CSA_RPC_FMT = rdf
  case *CSAS *CSAS.* *CSAS-*		# RSS 1.0 response mode
       CSA_PGM = CSA
       CSA_RPC_FMT = rss
  case *CSA2				# XML-RPC method call
       CSA_PGM = CSA2
       CSA_RPC_FMT = rdf
       CSA_XMLNS_DEFAULT = http://www.xmlrpc.com/spec
  case *CSAP				# SOAP method call
       CSA_PGM = CSAP
       CSA_RPC_FMT = rdf
  case *
       CSA_PGM = CSA			# default
}

CSA_ISINDEX=()					# Just in case

# Use special care on TMPDIR.
# The following mkdir+chmod sequence is subject to race conditions, but
# that isn't a problem, as it doesn't matter which process is doing
# what, provided that the directory is succesfully created.

if (!~ $TMPDIR^* $TMPDIR) {
   mkdir $TMPDIR && chmod 700 $TMPDIR || csaExit.fault 550
}

# Make sure errors are collected somewhere even before CSA_ROOT
# is set by the underlying application program. This is quite
# useful also to catch early errors that may occur before the
# application layer is initialized.

~ $CSA_DUMP_FILE /* || exec >>[2] $TMPDIR/stderr.log

# Perform some basic checks. To date, the CSA non-standard PUT and
# DELETE are supported, while their real HTTP counterparts are
# theoretically supported but have not been tested yet. Standard Web
# browsers generally only support GET/POST but programmatic clients may
# also use PUT/DELETE. Although DELETE does not require any content
# and could therefore be dispatched via GET, it would violate GET's
# idempotency property, thus POST is used instead (although possibly
# with Content-Length=0). The only form of QUERY_STRING supported by
# CSA in REST-mode is ISINDEX, therefore any other GET arguments must
# be specified in the form of PATH_INFO positional parameters. That is,
# rather than using "http://my.server/search?q=foo+bar", we will have
# to use something "http://my.server/search/foo+bar". Likewise, to
# specify that we want only from result item no. 10 onwards, we can use:
# "http://my.server/search/foo+bar/10". The thttpd web server, and maybe
# others too, supports the real HEAD method only for static web pages,
# not for CGI programs (for which a 501 status code is returned), so
# there is no way to have CSA handle a real HEAD request under thttpd
# because it wil never see the request.

switch ($REQUEST_METHOD) {
  case GET HEAD DELETE PUT;
  case POST
       if (~ $1 PUT DELETE) {
	  REQUEST_METHOD = $1
	  # Remove the REST verb, to prevent it from being
	  # treated as an argument by underlying CSA layers.
	  ~ $QUERY_STRING $REQUEST_METHOD && QUERY_STRING = ()
       }
  case *
       csaExit.fault 501
}

# Allocate the necessary work-files.

CSA_RPC_CMD = $TMPDIR/rpc$pid.tmp
CSA_RPC_AWK = $TMPDIR/awk$pid.tmp
CSA_RDF_ARG = $TMPDIR/arg$pid.tmp
CSA_RPC_PUT = $TMPDIR/put$pid.tmp
CSA_CGI_STDIN = $TMPDIR/stdin$pid.tmp

# Build a temporary AWK function cache if main file is missing.
# User-level overides cannot be loaded here, as we don't know
# where CSA_ROOT is yet, so we need to tell the main rc(1)
# library to also build the cache with all available functions.

if (~ $TMPDIR/fncache.awk* *'*' || ~ $CSA_AWK_NEWCACHE 1) {
   sed -n '/^[ 	]*\(#\|$\)/!{;s/^[ 	]*//;p;}' \
	$CSA_INSTALL/lib/include/*.awk > $TMPDIR/fncache$pid.tmp
   mv $TMPDIR/fncache$pid.tmp $TMPDIR/fncache.awk
   CSA_AWK_NEWCACHE = 1
}

# Warning: library load order *does* matter, at least with mawk(1),
# or really hard-to-trace bugs may occur !!

cmd = (_awk ''''$SWU_AWK_OPTS''''
    -f$SWU_INSTALL/lib/BEGIN.awk -f$CSA_INSTALL/lib/BEGIN.awk
    -f$SWU_INSTALL/lib/swulib.awk -f$TMPDIR/fncache.awk)

cmd = ($cmd -f^($CSA_INSTALL/lib/csa^((-rpciolib Cgi)^.awk))
						'''-v_PID='''$pid)

~ $CSA_PGM CSA[2P] && cmd = ($cmd '''-vRS=<''')

# Note the use of 'csa-escape -z' below. We need to choose an escaping
# mechanism that is unlikely to apply to part headers of multipart input.
# That is, we cannot use quoted-printable escaping, as '=' signs appare
# in most MIME defs in the input stream, and we would then need to un-escape
# them during multipart processing in csaCgi.awk.
# 
# There is a bug in MSIE 5.0x (and possibly other MSIE versions),
# whereby the browser does not include the trailing CR+LF in the
# content-length value that is supplied to the server. Because of this,
# the CGI program needs to read no more than 'content-length' bytes and
# then close the stdin stream. If this is not done, the server hangs
# and so does the client, until eventually the web server times-out
# and closes the connection and causing the operation to fail. This
# happens with thttpd/2.24, I have not tested it with other servers.
# CGI programs should then read no more than CONTENT_LENGTH bytes from
# stdin, and this is done internally by 'csa-escape -z' when used,
# while elsewhere it must be done by 'head -c' or other similar means.
# Truncating to CONTENT_LENGTH bytes has also the beneficial effect of
# dropping the trailing CR+LF supplied by MSIE, as well as any other
# trailing garbage which is not part of the payload. Note, however,
# that truncating may not always fix the problem, at least with thttpd,
# as the server buffers the data before sending it to the CGI program
# and the latter may not actually see the last input block until thttpd
# decides to flush it. See post_post_garbage_hack() in libthttpd.c.

# Any content-type is acceptable with requests which have no content.
if (!~ $REQUEST_METHOD GET HEAD) {

   switch ($CONTENT_TYPE) {

     case ();	# Void action, also for the HEAD/DELETE requests.

     case application/x-www-form-urlencoded*;
	  # No binary data are expected with this encoding type.

     case multipart/form-data*;

     case text/xml 'text/xml; '*

	  # Unless is PUT, text/xml can be accepted only
	  # with POST and either XML-RPC or SOAP.

	  ~ $REQUEST_METHOD PUT ||
	    { ~ $REQUEST_METHOD POST && ~ $CSA_PGM CSA[2P]} ||
	    csaExit.fault 406

     case *
	  # Any other MIME types can be accepted only with PUT.
	  ~ $REQUEST_METHOD PUT || csaExit.fault 406
   }
}

# If any input is present, then to keep on the safe side I always
# apply binary escaping, even if not all input MIME types are
# supposed to convey binary data.

# Note: opening stdin with no input causes some setups to hang.
# This happens for instance with thttpd/2.x.

if (~ $CONTENT_LENGTH [1-9]*) {
   #cmd = (head -c $CONTENT_LENGTH '|' csa-escape -z 
   if (~ $CSA_PGM(1) CSAP) {
      cmd = (csa-escape -z '|' _awk ''''$SWU_AWK_OPTS''''
			-f$CSA_INSTALL/lib/csa-soap2rpc.awk
			'>' $CSA_CGI_STDIN ';' $cmd -- $CSA_CGI_STDIN)
   }  else {
      cmd = (csa-escape -z '>' $CSA_CGI_STDIN ';' $cmd -- $CSA_CGI_STDIN)
   }
}  else {
   touch $CSA_CGI_STDIN				# must exist!
   cmd = ($cmd -- /dev/null)
}

eval $cmd || csaExit.fault

# Keep the environment as small as possible.
logname = ()
libstem = ()
tmpdir = ()
sessid = ()

. $CSA_RPC_CMD

csaExit $status					# This won't hurt.

# EOF
